import datetime
import os
from pathlib import Path
from unittest import mock

from pytest import fixture

from phylodata.data_types import (
    Experiment,
    ExperimentType,
    File,
    FileType,
    Metadata,
    Paper,
)
from phylodata.loader.preview_env import PREFER_PREVIEW_ENV
from phylodata.paper_with_experiment import PaperWithExperiment


@fixture()
def set_preview_env_var_to_true(monkeypatch):
    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            PREFER_PREVIEW_ENV: "true",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield


@fixture()
def set_preview_env_var_to_false(monkeypatch):
    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            PREFER_PREVIEW_ENV: "false",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield


@fixture
def experiment() -> PaperWithExperiment:
    return PaperWithExperiment(
        files=[
            File(
                name="beast2.xml",
                type=FileType.BEAST2_CONFIGURATION,
                size_bytes=0,
                md5="",
                local_path=Path("beast2.xml"),
            ),
            File(
                name="beast2 (preview).xml",
                type=FileType.BEAST2_CONFIGURATION,
                size_bytes=0,
                md5="",
                local_path=Path("beast2 (preview).xml"),
                is_preview=True,
            ),
            File(
                name="beast2.logs",
                type=FileType.BEAST2_POSTERIOR_LOGS,
                size_bytes=0,
                md5="",
                local_path=Path("beast2.logs"),
            ),
            File(
                name="beast2 (preview).logs",
                type=FileType.BEAST2_POSTERIOR_LOGS,
                size_bytes=0,
                md5="",
                local_path=Path("beast2 (preview).logs"),
                is_preview=True,
            ),
            File(
                name="beast2 (preview).trees",
                type=FileType.POSTERIOR_TREES,
                size_bytes=0,
                md5="",
                local_path=Path("beast2 (preview).trees"),
                is_preview=True,
            ),
        ],
        paper=Paper(
            doi="",
            title="",
            year=2018,
            authors=[],
            abstract="",
            bibtex="",
        ),
        experiment=Experiment(
            type=ExperimentType.BEAST2_Experiment,
            human_readable_id="",
            origin="",
            upload_date=datetime.date.today(),
            version=0,
            title="",
            description="",
        ),
        samples=[],
        evolutionary_model=[],
        trees=None,
        metadata=Metadata(evo_data_pipeline_version=""),
        local_path=Path(),
    )


def test_full_files_are_preferred_if_nothing_specified(experiment: PaperWithExperiment):
    files = experiment.get_files()
    assert len(files) == 3
    assert files[0].name == "beast2.xml"
    assert files[1].name == "beast2.logs"
    assert files[2].name == "beast2 (preview).trees"


def test_only_full_files_are_returned_if_preview_is_false(
    experiment: PaperWithExperiment,
):
    files = experiment.get_files(prefer_preview=False)
    assert len(files) == 2
    assert files[0].name == "beast2.xml"
    assert files[1].name == "beast2.logs"


def test_previews_are_preferred_if_preview_is_true(
    experiment: PaperWithExperiment,
):
    files = experiment.get_files(prefer_preview=True)
    assert len(files) == 3
    assert files[0].name == "beast2 (preview).xml"
    assert files[1].name == "beast2 (preview).logs"
    assert files[2].name == "beast2 (preview).trees"



import os
from collections import defaultdict
from pathlib import Path
from typing import Optional

from phylodata.data_types import (
    EditablePaperWithExperiment,
    EvolutionaryModelComponent,
    Experiment,
    File,
    FileType,
    Metadata,
    NonEditablePaperWithExperiment,
    Paper,
    Sample,
    Trees,
    msgspec,
)
from phylodata.loader.preview_env import PREFER_PREVIEW_ENV


class PaperWithExperiment(msgspec.Struct, rename="camel"):
    paper: Paper
    experiment: Experiment
    files: list[File]
    samples: list[Sample]
    trees: Optional[Trees]
    evolutionary_model: list[EvolutionaryModelComponent]
    metadata: Metadata
    local_path: Path

    @classmethod
    def from_partial(
        cls,
        editable: EditablePaperWithExperiment,
        non_editable: NonEditablePaperWithExperiment,
        local_path: Path,
    ):
        return cls(
            paper=Paper.from_partial(editable.paper, non_editable.paper),
            experiment=Experiment.from_partial(
                editable.experiment, non_editable.experiment
            ),
            files=non_editable.files,
            samples=editable.samples,
            trees=non_editable.trees,
            evolutionary_model=non_editable.evolutionary_model,
            metadata=non_editable.metadata,
            local_path=local_path,
        )

    def get_files(
        self,
        prefer_preview: Optional[bool] = None,
    ) -> list[File]:
        """Retrieves all downloaded files of the experiment.
        For every file, either the full or preview version is returned.

        Args:
            prefer_preview:
                If None, prefers a full file but falls back to a preview file if no full file exists.
                If True, prefers and returns a preview file if available.
                If False, only retrieves full files.
                This can also be controlled by setting the environment variable PHYLODATA_PREFER_PREVIEW
                to true or false.

        Returns:
            A list of File objects containing the experiment data.
        """
        if prefer_preview is None and os.environ.get(PREFER_PREVIEW_ENV) is not None:
            prefer_preview = os.environ.get(PREFER_PREVIEW_ENV) == "true"

        downloaded_files = [file for file in self.files if file.local_path]

        downloaded_file_variants = defaultdict(list[File])
        for file in downloaded_files:
            downloaded_file_variants[self._strip_preview_suffix(file)].append(file)

        result = []
        for variants in downloaded_file_variants.values():
            preview_variant = [file for file in variants if file.is_preview]
            full_variant = [file for file in variants if not file.is_preview]

            if prefer_preview is None:
                # Prefer full, fall back to preview
                if full_variant:
                    result.extend(full_variant)
                elif preview_variant:
                    result.extend(preview_variant)

            elif prefer_preview:
                # Prefer preview, fall back to full
                if preview_variant:
                    result.extend(preview_variant)
                elif full_variant:
                    result.extend(full_variant)

            else:
                # Only full, skip if none exists
                if full_variant:
                    result.extend(full_variant)

        return result

    def _strip_preview_suffix(self, file: File) -> str:
        name = file.name
        extension = name.split(".")[-1]

        if name.endswith(f" (preview).{extension}"):
            return name[: -len(" (preview)") - len(extension) - 1] + "." + extension

        return name

    def get_files_of_type(
        self,
        file_type: FileType,
        prefer_preview: Optional[bool] = None,
    ) -> list[File]:
        """Retrieves all downloaded files of the experiment of a specific type.
        For every file, either the full or preview version is returned.

        Args:
            file_type:
                The FileType to retrieve.
            prefer_preview:
                If None, prefers a full file but falls back to a preview file if no full file exists.
                If True, prefers and returns a preview file if available.
                If False, only retrieves full files.
                This can also be controlled by setting the environment variable PHYLODATA_PREFER_PREVIEW
                to true or false.

        Returns:
            A list of File objects containing the experiment data.
        """
        return [
            file for file in self.get_files(prefer_preview) if file.type == file_type
        ]

    def get_file(
        self,
        name: str,
        prefer_preview: Optional[bool] = None,
    ) -> Optional[File]:
        """Retrieves a file of a specific name from a PaperWithExperiment object.

        Args:
            name:
                The name of the file to retrieve.
            prefer_preview:
                If None, prefers a full file but falls back to a preview file if no full file exists.
                If True, prefers and returns a preview file if available.
                If False, only retrieves full files.
                This can also be controlled by setting the environment variable PHYLODATA_PREFER_PREVIEW
                to true or false.

        Returns:
            The first matching File object, or None if no such file exists.
        Raises:
            FileNotFoundError if no matching file is found.
        """
        matching_files = [
            file
            for file in self.get_files(prefer_preview)
            if self._strip_preview_suffix(file) == name
        ]

        if not matching_files:
            raise FileNotFoundError

        return matching_files[0]

    def get_file_of_type(
        self,
        file_type: FileType,
        prefer_preview: Optional[bool] = None,
    ) -> Optional[File]:
        """
        Retrieves a file of a specified type from a PaperWithExperiment object.

        Args:
            file_type:
                The FileType to retrieve.
            prefer_preview:
                If None, prefers a full file but falls back to a preview file if no full file exists.
                If True, prefers and returns a preview file if available.
                If False, always retrieves a full file.
                This can also be controlled by setting the environment variable PHYLODATA_PREFER_PREVIEW
                to true or false.

        Returns:
            The first matching File object, or None if no such file exists.
        Raises:
            FileNotFoundError if no matching file is found.
        """
        matching_files = self.get_files_of_type(file_type, prefer_preview)

        if not matching_files:
            raise FileNotFoundError

        return matching_files[0]



import datetime
import os
from pathlib import Path
from unittest import mock

from pytest import fixture

from phylodata.data_types import (
    Experiment,
    ExperimentType,
    File,
    FileType,
    Metadata,
    Paper,
)
import pytest
from phylodata.loader.preview_env import PREFER_PREVIEW_ENV
from phylodata.paper_with_experiment import PaperWithExperiment


@fixture()
def set_preview_env_var_to_true(monkeypatch):
    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            PREFER_PREVIEW_ENV: "true",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield


@fixture()
def set_preview_env_var_to_false(monkeypatch):
    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            PREFER_PREVIEW_ENV: "false",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield


@fixture
def experiment() -> PaperWithExperiment:
    return PaperWithExperiment(
        files=[],
        paper=Paper(
            doi="",
            title="",
            year=2018,
            authors=[],
            abstract="",
            bibtex="",
        ),
        experiment=Experiment(
            type=ExperimentType.BEAST2_Experiment,
            human_readable_id="",
            origin="",
            upload_date=datetime.date.today(),
            version=0,
            title="",
            description="",
        ),
        samples=[],
        evolutionary_model=[],
        trees=None,
        metadata=Metadata(evo_data_pipeline_version=""),
        local_path=Path(),
    )


def test_single_possible_file_get_found(experiment: PaperWithExperiment):
    experiment.files = [
        File(
            name="beast.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path(),
        ),
        File(
            name="beast.trees",
            type=FileType.POSTERIOR_TREES,
            size_bytes=0,
            md5="",
            local_path=Path(),
        ),
        File(
            name="summary.trees",
            type=FileType.SUMMARY_TREE,
            size_bytes=0,
            md5="",
            local_path=Path(),
        ),
    ]

    config = experiment.get_file_of_type(FileType.BEAST2_CONFIGURATION)
    assert config
    assert config.name == "beast.xml"

    posterior = experiment.get_file_of_type(FileType.POSTERIOR_TREES)
    assert posterior
    assert posterior.name == "beast.trees"

    summary = experiment.get_file_of_type(FileType.SUMMARY_TREE)
    assert summary
    assert summary.name == "summary.trees"


def test_missing_file_type_raises_error(experiment: PaperWithExperiment):
    experiment.files = [
        File(
            name="beast.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path(),
        ),
    ]

    with pytest.raises(FileNotFoundError):  # noqa: F821
        experiment.get_file_of_type(FileType.SUMMARY_TREE)


def test_multiple_possible_files_returns_first(experiment: PaperWithExperiment):
    experiment.files = [
        File(
            name="beast.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path(),
        ),
        File(
            name="beast2.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path(),
        ),
    ]

    config = experiment.get_file_of_type(FileType.BEAST2_CONFIGURATION)
    assert config
    assert config.name == "beast.xml"


def test_full_files_are_preferred_if_nothing_specified(experiment: PaperWithExperiment):
    experiment.files = [
        File(
            name="beast.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast.xml"),
        ),
        File(
            name="beast2 (preview).xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast2 (preview).xml"),
            is_preview=True,
        ),
        File(
            name="trees.xml",
            type=FileType.POSTERIOR_TREES,
            size_bytes=0,
            md5="",
            local_path=Path("trees.xml"),
        ),
        File(
            name="logs (preview).xml",
            type=FileType.BEAST2_POSTERIOR_LOGS,
            size_bytes=0,
            md5="",
            local_path=Path("logs (preview).xml"),
            is_preview=True,
        ),
    ]

    config = experiment.get_file_of_type(FileType.BEAST2_CONFIGURATION)
    assert config
    assert config.name == "beast.xml"

    posterior = experiment.get_file_of_type(FileType.POSTERIOR_TREES)
    assert posterior
    assert posterior.name == "trees.xml"

    logs = experiment.get_file_of_type(FileType.BEAST2_POSTERIOR_LOGS)
    assert logs
    assert logs.name == "logs (preview).xml"


def test_only_full_files_are_returned_if_preview_is_false(
    experiment: PaperWithExperiment,
):
    experiment.files = [
        File(
            name="beast.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast.xml"),
        ),
        File(
            name="beast2 (preview).xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast2 (preview).xml"),
            is_preview=True,
        ),
        File(
            name="trees.xml",
            type=FileType.POSTERIOR_TREES,
            size_bytes=0,
            md5="",
            local_path=Path("trees.xml"),
        ),
        File(
            name="logs (preview).xml",
            type=FileType.BEAST2_POSTERIOR_LOGS,
            size_bytes=0,
            md5="",
            local_path=Path("logs (preview).xml"),
            is_preview=True,
        ),
    ]

    config = experiment.get_file_of_type(
        FileType.BEAST2_CONFIGURATION, prefer_preview=False
    )
    assert config
    assert config.name == "beast.xml"

    posterior = experiment.get_file_of_type(
        FileType.POSTERIOR_TREES, prefer_preview=False
    )
    assert posterior
    assert posterior.name == "trees.xml"

    with pytest.raises(FileNotFoundError):
        experiment.get_file_of_type(
            FileType.BEAST2_POSTERIOR_LOGS, prefer_preview=False
        )


def test_previews_are_preferred_if_preview_is_true(
    experiment: PaperWithExperiment,
):
    experiment.files = [
        File(
            name="beast2.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast2.xml"),
        ),
        File(
            name="beast2 (preview).xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast2 (preview).xml"),
            is_preview=True,
        ),
        File(
            name="trees.xml",
            type=FileType.POSTERIOR_TREES,
            size_bytes=0,
            md5="",
            local_path=Path("trees.xml"),
        ),
        File(
            name="logs (preview).xml",
            type=FileType.BEAST2_POSTERIOR_LOGS,
            size_bytes=0,
            md5="",
            local_path=Path("logs (preview).xml"),
            is_preview=True,
        ),
    ]

    config = experiment.get_file_of_type(
        FileType.BEAST2_CONFIGURATION, prefer_preview=True
    )
    assert config
    assert config.name == "beast2 (preview).xml"

    posterior = experiment.get_file_of_type(
        FileType.POSTERIOR_TREES, prefer_preview=True
    )
    assert posterior
    assert posterior.name == "trees.xml"

    logs = experiment.get_file_of_type(
        FileType.BEAST2_POSTERIOR_LOGS, prefer_preview=True
    )
    assert logs
    assert logs.name == "logs (preview).xml"


def test_previews_are_preferred_if_preview_env_is_set_true(
    experiment: PaperWithExperiment, set_preview_env_var_to_true
):
    experiment.files = [
        File(
            name="beast2.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast2.xml"),
        ),
        File(
            name="beast2 (preview).xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast2 (preview).xml"),
            is_preview=True,
        ),
    ]

    config = experiment.get_file_of_type(FileType.BEAST2_CONFIGURATION)
    assert config
    assert config.name == "beast2 (preview).xml"


def test_previews_are_preferred_if_preview_env_is_set_false(
    experiment: PaperWithExperiment, set_preview_env_var_to_false
):
    experiment.files = [
        File(
            name="beast2.xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast2.xml"),
        ),
        File(
            name="beast2 (preview).xml",
            type=FileType.BEAST2_CONFIGURATION,
            size_bytes=0,
            md5="",
            local_path=Path("beast2 (preview).xml"),
            is_preview=True,
        ),
    ]

    config = experiment.get_file_of_type(FileType.BEAST2_CONFIGURATION)
    assert config
    assert config.name == "beast2.xml"



import os

PREFER_PREVIEW_ENV = "PHYLODATA_PREFER_PREVIEW"


def prefer_preview():
    """Sets the PHYLODATA_PREFER_PREVIEW environment variable to true."""
    os.environ[PREFER_PREVIEW_ENV] = "true"


def prefer_full():
    """Sets the PHYLODATA_PREFER_PREVIEW environment variable to false."""
    os.environ[PREFER_PREVIEW_ENV] = "false"



import os
from collections import defaultdict
from pathlib import Path
from typing import Optional

from phylodata.data_types import File, FileType
from phylodata.loader.preview_env import PREFER_PREVIEW_ENV
from phylodata.paper_with_experiment import PaperWithExperiment


def get_folder(experiment: PaperWithExperiment) -> Path:
    """Returns the folder containing the local experiment files."""
    return next(
        (file.local_path.parent for file in experiment.files if file.local_path),
        Path("data") / experiment.experiment.human_readable_id,
    )


def get_files(
    experiment: PaperWithExperiment,
    prefer_preview: Optional[bool] = None,
) -> list[File]:
    """Retrieves all downloaded files of the experiment.
    For every file, either the full or preview version is returned.

    Args:
        experiment:
            The experiement to retrieve files from.
        prefer_preview:
            If None, prefers a full file but falls back to a preview file if no full file exists.
            If True, prefers and returns a preview file if available.
            If False, only retrieves full files.
            This can also be controlled by setting the environment variable PHYLODATA_PREFER_PREVIEW
            to true or false.

    Returns:
        A list of File objects containing the experiment data.
    """
    if prefer_preview is None and os.environ.get(PREFER_PREVIEW_ENV) is not None:
        prefer_preview = os.environ.get(PREFER_PREVIEW_ENV) == "true"

    downloaded_files = [file for file in experiment.files if file.local_path]

    downloaded_file_variants = defaultdict(list[File])
    for file in downloaded_files:
        downloaded_file_variants[_strip_preview_suffix(file)].append(file)

    result = []
    for variants in downloaded_file_variants.values():
        preview_variant = [file for file in variants if file.is_preview]
        full_variant = [file for file in variants if not file.is_preview]

        if prefer_preview is None:
            # Prefer full, fall back to preview
            if full_variant:
                result.extend(full_variant)
            elif preview_variant:
                result.extend(preview_variant)

        elif prefer_preview:
            # Prefer preview, fall back to full
            if preview_variant:
                result.extend(preview_variant)
            elif full_variant:
                result.extend(full_variant)

        else:
            # Only full, skip if none exists
            if full_variant:
                result.extend(full_variant)

    return result


def _strip_preview_suffix(file: File) -> str:
    name = file.name
    extension = name.split(".")[-1]

    if name.endswith(f" (preview).{extension}"):
        return name[: -len(" (preview)") - len(extension) - 1] + "." + extension

    return name


def get_files_of_type(
    experiment: PaperWithExperiment,
    file_type: FileType,
    prefer_preview: Optional[bool] = None,
) -> list[File]:
    """Retrieves all downloaded files of the experiment of a specific type.
    For every file, either the full or preview version is returned.

    Args:
        experiment:
            The experiement to retrieve files from.
        file_type:
            The FileType to retrieve.
        prefer_preview:
            If None, prefers a full file but falls back to a preview file if no full file exists.
            If True, prefers and returns a preview file if available.
            If False, only retrieves full files.
            This can also be controlled by setting the environment variable PHYLODATA_PREFER_PREVIEW
            to true or false.

    Returns:
        A list of File objects containing the experiment data.
    """
    return [
        file for file in get_files(experiment, prefer_preview) if file.type == file_type
    ]


def get_file(
    experiment: PaperWithExperiment,
    name: str,
    prefer_preview: Optional[bool] = None,
) -> Optional[File]:
    """Retrieves a file of a specific name from a PaperWithExperiment object.

    Args:
        experiment:
            The PaperWithExperiment object containing the files.
        name:
            The name of the file to retrieve.
        prefer_preview:
            If None, prefers a full file but falls back to a preview file if no full file exists.
            If True, prefers and returns a preview file if available.
            If False, only retrieves full files.
            This can also be controlled by setting the environment variable PHYLODATA_PREFER_PREVIEW
            to true or false.

    Returns:
        The first matching File object, or None if no such file exists.
    """
    matching_files = [
        file
        for file in get_files(experiment, prefer_preview)
        if _strip_preview_suffix(file) == name
    ]
    return matching_files[0] if matching_files else None


def get_file_of_type(
    experiment: PaperWithExperiment,
    file_type: FileType,
    prefer_preview: Optional[bool] = None,
) -> Optional[File]:
    """
    Retrieves a file of a specified type from a PaperWithExperiment object.

    Args:
        experiment:
            The PaperWithExperiment object containing the files.
        file_type:
            The FileType to retrieve.
        prefer_preview:
            If None, prefers a full file but falls back to a preview file if no full file exists.
            If True, prefers and returns a preview file if available.
            If False, always retrieves a full file.
            This can also be controlled by setting the environment variable PHYLODATA_PREFER_PREVIEW
            to true or false.

    Returns:
        The first matching File object, or None if no such file exists.
    """
    matching_files = get_files_of_type(experiment, file_type, prefer_preview)
    return matching_files[0] if matching_files else None



from pathlib import Path
from typing import Optional

import msgspec

from phylodata.data_types import ClassificationEntryType, EditablePaperWithExperiment


def remove_classification(
    metadata_file: Path,
    sample_id: Optional[str] = None,
    classification_type: Optional[ClassificationEntryType] = None,
):
    """
    Removes the classification of a language sample in a editable metadata file.
    If no sample_id is provided, all samples are removed.
    If classification_type is given, all samples with that classification type are removed.

    Args:
        metadata_file: The path to the metadata file.
        sample_id: The ID of the sample to remove the classification from.
        classification_type: The type of classifications to remove.
    """

    with open(metadata_file, "r") as file:
        metadata = msgspec.json.decode(file.read(), type=EditablePaperWithExperiment)

    for sample in metadata.samples:
        if sample_id and sample.sample_id != sample_id:
            continue

        if classification_type and classification_type not in {
            data.id_type for data in sample.classification
        }:
            continue

        sample.classification = []
        sample.scientific_name = sample.sample_id
        sample.common_name = None

    with open(metadata_file, "wb") as file:
        file.write(msgspec.json.format(msgspec.json.encode(metadata), indent=2))
